module RedBlack where

open import Prelude

data Color : Set where
  Red : Color
  Blk : Color

 -- Red-black tree.
 -- The second argument records the color of the root.

data RBTree (A : Set) : Color → ℕ → Set where
  E : RBTree A Blk 0
  R : ∀ {n}       → RBTree A Blk n → A → RBTree A Blk n → RBTree A Red n
  B : ∀ {n cₗ cᵣ} → RBTree A cₗ   n → A → RBTree A cᵣ   n → RBTree A Blk (suc n)

 -- "Infrared" tree: almost a Red-rooted red-black tree, but we allow
 -- the red root to have one red child.
 --   RL -- the left child is red.
 --   RR -- the right child is red.
 --   RB -- both children are black -- that is, a regular red-black tree.

data InfRedTree (A : Set) : ℕ → Set where
  RL : ∀ {n}   → RBTree A Red n → A → RBTree A Blk n → InfRedTree A n
  RR : ∀ {n}   → RBTree A Blk n → A → RBTree A Red n → InfRedTree A n
  RB : ∀ {n}   → RBTree A Blk n → A → RBTree A Blk n → InfRedTree A n

 -- a smart constructor that generates an InfRedTree given two 
 -- subtrees, at least one of them being black.

mkR : ∀ {A n} → (ct : ∃ (λ c → RBTree A c n)) → A → (cu : ∃ (λ c → RBTree A c n))
    → (fst ct ≡ Blk) ⊎ (fst cu ≡ Blk) -- determining which child is black
    → InfRedTree A n
mkR (Red , t) x (Red , u) (inl ())
mkR (Red , t) x (Red , u) (inr ())
mkR (Red , t) x (Blk , u) _ = RL t x u
mkR (Blk , t) x (Red , u) _ = RR t x u
mkR (Blk , t) x (Blk , u) _ = RB t x u

data Order : Set where
  LT : Order
  EQ : Order
  GT : Order

  -- we abstract away the element type A and its comparison 
  -- operator, since they do not matter in this example.

postulate 
  A : Set
  cmp : A → A → Order

  -- migtBeRBT c is either InfRedTree or RBTree, depending on c
  -- We need it in the type of ins.

MigtBeRBT : Color → Set → ℕ → Set
MigtBeRBT Red A n = InfRedTree A n
MigtBeRBT Blk A n = ∃ (λ c → RBTree A c n)

-- Now we may define the function ins, which calls balance. 
-- For readability, we use a "mutual" keyword to allow ins
-- to be defined before balance(L/R), even though they are
-- not mutually recursive.

mutual  
      
 ins : ∀ {c n} → A → RBTree A c n → {!   !}
 ins k E = {!   !}
 ins {n} k (R t x u) with cmp k x 
 ... | LT = {!   !} -- mkR might be useful here
 ... | EQ = {!   !}
 ... | GT = {!   !}
 ins k (B t x u) with cmp k x
 ... | LT = {!   !} -- might call balanceL here
 ... | EQ = {!   !}
 ... | GT = {!   !} -- might call balanceR here

  -- the function balance in the Haskell version is split into two
  -- functions, according to which argument is a result of ins. 

 balanceL : {cₗ cᵣ : Color} {n : ℕ} → {!   !} → A → {!   !} → {!   !}
 balanceL {cₗ} t x u = {!   !}

 balanceR : {cₗ cᵣ : Color} {n : ℕ} → {!   !} → A → {!   !} → {!   !}
 balanceR {_} {cᵣ} t x u = {!   !}


  -- we do not know the height of the tree after insertion,
  -- therefore it might be apporpriate to make insert return a sum type:
  --  insert : ... → something ⊎ something

insert : ∀ {n} → A → RBTree A Blk n → {!   !}
insert {n} k t = blacken (ins k t)
  where blacken : {!   !} → {!   !}
        blacken = {!   !}
